// ==============================================================
//	This file is part of Glest (www.glest.org)
//
//	Copyright (C) 2001-2005 Martio Figueroa
//
//	You can redistribute this code and/or modify it under 
//	the terms of the GNU General Public License as published 
//	by the Free Software Foundation; either version 2 of the 
//	License, or (at your option) any later version
// ==============================================================

#include "main_menu.h" 

#include "renderer.h"
#include "sound.h"
#include "config.h"
#include "program.h"
#include "game_util.h" 
#include "game.h"
#include "platform_util.h"
#include "sound_renderer.h"
#include "core_data.h"
#include "faction.h"
#include "metrics.h"
#include "leak_dumper.h"

using namespace Shared::Sound;
using namespace Shared::Util;
using namespace Shared::Graphics;
using namespace	Shared::Xml;

namespace Glest{ namespace Game{

// =====================================================
// 	class MainMenu
// =====================================================

// ===================== PUBLIC ========================

MainMenu::MainMenu(Program *program):
	ProgramState(program)
{
	mouseX=100; 
	mouseY=100;

	state= NULL;
	this->program= program;

	fps= 0;
	lastFps= 0;

	setState(new MenuStateRoot(program, this));
	SoundRenderer &soundRenderer= SoundRenderer::getInstance();
}

MainMenu::~MainMenu(){
	delete state;
	Renderer::getInstance().endMenu();
	SoundRenderer &soundRenderer= SoundRenderer::getInstance();
	soundRenderer.stopAllSounds();
};

void MainMenu::init(){
	Renderer::getInstance().initMenu(this);
}

//asynchronus render update
void MainMenu::render(){

	Config &config= Config::getInstance();
	Renderer &renderer= Renderer::getInstance();
	CoreData &coreData= CoreData::getInstance();

	fps++;

    renderer.clearBuffers();
	
	//3d
	renderer.reset3dMenu();
	renderer.clearZBuffer();
	renderer.loadCameraMatrix(menuBackground.getCamera());
	renderer.renderMenuBackground(&menuBackground);
	renderer.renderParticleManager(rsMenu);

	//2d
	renderer.reset2d();
	state->render();
    renderer.renderMouse2d(mouseX, mouseY, mouse2dAnim);

    if(config.getBool("DebugMode")){
		renderer.renderText(
			"FPS: " + intToStr(lastFps), 
			coreData.getMenuFontNormal(), Vec3f(1.f), 10, 10, false);
    }

	renderer.swapBuffers();
}

//syncronus update
void MainMenu::update(){
	Renderer::getInstance().updateParticleManager(rsMenu);
	mouse2dAnim= (mouse2dAnim +1) % Renderer::maxMouse2dAnim;
	menuBackground.update();
}

void MainMenu::tick(){
	lastFps= fps;
	fps= 0;
}

//event magangement: mouse click
void MainMenu::mouseMove(int x, int y, const MouseState *ms){
    mouseX= x; mouseY= y;
	state->mouseMove(x, y, ms);
}

//returns if exiting
void MainMenu::mouseDownLeft(int x, int y){
	state->mouseClick(x, y, mbLeft);	
}

void MainMenu::mouseDownRight(int x, int y){
	state->mouseClick(x, y, mbRight);	
}

void MainMenu::setState(MenuState *state){
	delete this->state;
	this->state= state;
	GraphicComponent::resetFade();

	menuBackground.setTargetCamera(state->getCamera());
}


// =====================================================
// 	class MenuState
// =====================================================

MenuState::MenuState(Program *program, MainMenu *mainMenu, const string &stateName){
	this->program= program;
	this->mainMenu= mainMenu;

	//camera
	XmlTree xmlTree;
	xmlTree.load("data/core/menu/menu.xml");
	const XmlNode *menuNode= xmlTree.getRootNode();
	const XmlNode *cameraNode= menuNode->getChild("camera");

	//position
	const XmlNode *positionNode= cameraNode->getChild(stateName + "-position");
	Vec3f startPosition;
    startPosition.x= positionNode->getAttribute("x")->getFloatValue();
	startPosition.y= positionNode->getAttribute("y")->getFloatValue();
	startPosition.z= positionNode->getAttribute("z")->getFloatValue();
	camera.setPosition(startPosition);

	//rotation
	const XmlNode *rotationNode= cameraNode->getChild(stateName + "-rotation");
	Vec3f startRotation;
    startRotation.x= rotationNode->getAttribute("x")->getFloatValue();
	startRotation.y= rotationNode->getAttribute("y")->getFloatValue();
	startRotation.z= rotationNode->getAttribute("z")->getFloatValue();
	camera.setOrientation(Quaternion(EulerAngles(
		degToRad(startRotation.x), 
		degToRad(startRotation.y), 
		degToRad(startRotation.z))));
}


// =====================================================
// 	class MenuStateRoot
// =====================================================

MenuStateRoot::MenuStateRoot(Program *program, MainMenu *mainMenu): 
	MenuState(program, mainMenu, "root")
{
	Lang &lang= Lang::getInstance();

	buttonPlay.init(425, 350, 150);
    buttonConfig.init(425, 300, 150);
    buttonAbout.init(425, 250, 150);
    buttonExit.init(425, 200, 150);
	labelVersion.init(520, 440);

	buttonPlay.setText(lang.get("Play"));
	buttonConfig.setText(lang.get("Config"));
	buttonAbout.setText(lang.get("About")); 
	buttonExit.setText(lang.get("Exit"));
	labelVersion.setText(glestVersionString);

	CoreData &coreData= CoreData::getInstance();
	Font *font= coreData.getMenuFontNormal(); 
}

void MenuStateRoot::mouseClick(int x, int y, MouseButton mouseButton){

	CoreData &coreData=  CoreData::getInstance();
	SoundRenderer &soundRenderer= SoundRenderer::getInstance();

	if(buttonPlay.mouseClick(x, y)){
		soundRenderer.playFx(coreData.getClickSoundB());
		mainMenu->setState(new MenuStateGameSettings(program, mainMenu));
    }     
    else if(buttonConfig.mouseClick(x, y)){ 
		soundRenderer.playFx(coreData.getClickSoundB());
		mainMenu->setState(new MenuStateConfig(program, mainMenu));
    }
    else if(buttonAbout.mouseClick(x, y)){ 
		soundRenderer.playFx(coreData.getClickSoundB());
		mainMenu->setState(new MenuStateAbout(program, mainMenu));
    }
    else if(buttonExit.mouseClick(x, y)){
		soundRenderer.playFx(coreData.getClickSoundA());
		program->exit();
    }
}

void MenuStateRoot::mouseMove(int x, int y, const MouseState *ms){
	buttonPlay.mouseMove(x, y);
    buttonConfig.mouseMove(x, y);
    buttonAbout.mouseMove(x, y); 
    buttonExit.mouseMove(x,y);
}

void MenuStateRoot::render(){
	Renderer &renderer= Renderer::getInstance();
	CoreData &coreData= CoreData::getInstance();
	const Metrics &metrics= Metrics::getInstance();
	Font *font= coreData.getMenuFontNormal(); 
	
	int w= 300;
	int h= 150;

	renderer.renderTextureQuad(
		(metrics.getVirtualW()-w)/2, 475-h/2, w, h, 
		coreData.getLogoTexture(), GraphicComponent::getFade());
	renderer.renderButton(&buttonPlay);
	renderer.renderButton(&buttonConfig);
	renderer.renderButton(&buttonAbout);
	renderer.renderButton(&buttonExit);
	renderer.renderLabel(&labelVersion);
}


// =====================================================
// 	class MenuStateGameSettings
// =====================================================

MenuStateGameSettings::MenuStateGameSettings(Program *program, MainMenu *mainMenu): 
	MenuState(program, mainMenu, "settings")
{
	Lang &lang= Lang::getInstance();

	vector<string> results, teamItems, controlItems;

	//create
	buttonReturn.init(350, 200, 125);
	buttonPlayNow.init(525, 200, 125);

    //map listBox
    findAll("maps/*.gbm", results, true);
	if(results.size()==0){
        throw runtime_error("There is no map");
	}
	mapFiles= results;
	for(int i= 0; i<results.size(); ++i){
		results[i]= formatString(results[i]);
	}
    listBoxMap.init(200, 320);
    listBoxMap.setItems(results);
	labelMap.init(200, 350);
	labelMapInfo.init(200, 290, 200, 40);

    //tileset listBox
    findAll("tilesets/*.", results);
	if(results.size()==0){
        throw runtime_error("There is no tile set");
	}
    tilesetFiles= results;
	for(int i= 0; i<results.size(); ++i){
		results[i]= formatString(results[i]);
	}
	listBoxTileset.init(400, 320);
    listBoxTileset.setItems(results);
	labelTileset.init(400, 350);

    //tech Tree listBox
    findAll("techs/*.", results);
	if(results.size()==0){
        throw runtime_error("There is no tech tree");
	}
    techTreeFiles= results;
	for(int i= 0; i<results.size(); ++i){
		results[i]= formatString(results[i]);
	}
	listBoxTechTree.init(600, 320);
    listBoxTechTree.setItems(results);
	labelTechTree.init(600, 350);

	//list boxes
    for(int i=0; i<GameConstants::maxPlayers; ++i){
		labelPlayers[i].init(200, 500-i*30);
        listBoxControls[i].init(300, 500-i*30);
        listBoxFactions[i].init(500, 500-i*30);
		listBoxTeams[i].init(700, 500-i*30, 60);
    }

	labelControl.init(300, 550, GraphicListBox::defW, GraphicListBox::defH, true);
    labelFaction.init(500, 550, GraphicListBox::defW, GraphicListBox::defH, true);
    labelTeam.init(700, 550, 60, GraphicListBox::defH, true);

	//texts
	buttonReturn.setText(lang.get("Return"));
	buttonPlayNow.setText(lang.get("PlayNow"));

    controlItems.push_back(lang.get("Closed"));
	controlItems.push_back(lang.get("Cpu"));
    controlItems.push_back(lang.get("CpuUltra"));
	controlItems.push_back(lang.get("Human"));
	teamItems.push_back("1");
	teamItems.push_back("2");
	teamItems.push_back("3");
	teamItems.push_back("4");

	reloadFactions();

	findAll("techs/"+techTreeFiles[listBoxTechTree.getSelectedItemIndex()]+"/factions/*.", results);
	if(results.size()==0){
        throw runtime_error("There is no factions for this tech tree");
	}
   
	for(int i=0; i<GameConstants::maxPlayers; ++i){
		labelPlayers[i].setText(lang.get("Player")+" "+intToStr(i));
        listBoxTeams[i].setItems(teamItems);
		listBoxTeams[i].setSelectedItemIndex(i);
		listBoxControls[i].setItems(controlItems); 
    }

	labelMap.setText(lang.get("Map"));
	labelTileset.setText(lang.get("Tileset"));
	labelTechTree.setText(lang.get("TechTree"));
	labelControl.setText(lang.get("Control"));
    labelFaction.setText(lang.get("Faction"));
    labelTeam.setText(lang.get("Team"));

	loadMapInfo("maps/"+mapFiles[listBoxMap.getSelectedItemIndex()]+".gbm", &mapInfo);
	
	labelMapInfo.setText(mapInfo.desc);

	updateControlers();
}


void MenuStateGameSettings::mouseClick(int x, int y, MouseButton mouseButton){

	CoreData &coreData= CoreData::getInstance();
	SoundRenderer &soundRenderer= SoundRenderer::getInstance();

	if(buttonReturn.mouseClick(x,y)){
		soundRenderer.playFx(coreData.getClickSoundA());
		mainMenu->setState(new MenuStateRoot(program, mainMenu));
    }  
	else if(buttonPlayNow.mouseClick(x,y)){
		soundRenderer.playFx(coreData.getClickSoundC());
		GameSettings *gameSettings= new GameSettings();
		loadGameSettings(gameSettings);
		program->setState(new Game(program, gameSettings));
	}
	else if(listBoxMap.mouseClick(x, y)){
		loadMapInfo("maps/"+mapFiles[listBoxMap.getSelectedItemIndex()]+".gbm", &mapInfo);
		labelMapInfo.setText(mapInfo.desc);
		updateControlers();
	}
	else if(listBoxTileset.mouseClick(x, y)){
	}
	else if(listBoxTechTree.mouseClick(x, y)){
		reloadFactions();
	}
	else{
		for(int i=0; i<mapInfo.players; ++i){

			//ensure thet only 1 human player is present
			if(listBoxControls[i].mouseClick(x, y)){

				//look for human players
				int humanIndex1= -1;
				int humanIndex2= -1;
				for(int j=0; j<GameConstants::maxPlayers; ++j){
					ControlerType ct= static_cast<ControlerType>(listBoxControls[j].getSelectedItemIndex());
					if(ct==ctHuman){
						if(humanIndex1==-1){
							humanIndex1= j;
						}
						else{
							humanIndex2= j;
						}
					}
				}

				//no human
				if(humanIndex1==-1 && humanIndex2==-1){
					listBoxControls[i].setSelectedItemIndex(ctHuman);
				}

				//2 humans
				if(humanIndex1!=-1 && humanIndex2!=-1){
					listBoxControls[humanIndex1==i? humanIndex2: humanIndex1].setSelectedItemIndex(ctClosed);
				}
			}
			else if(listBoxFactions[i].mouseClick(x, y)){
			}
			else if(listBoxTeams[i].mouseClick(x, y)){
			}
		}
	}
}

void MenuStateGameSettings::mouseMove(int x, int y, const MouseState *ms){
	
	buttonReturn.mouseMove(x, y);
	buttonPlayNow.mouseMove(x, y);

	for(int i=0; i<GameConstants::maxPlayers; ++i){
        listBoxControls[i].mouseMove(x, y);
        listBoxFactions[i].mouseMove(x, y);
		listBoxTeams[i].mouseMove(x, y);
    }
	listBoxMap.mouseMove(x, y);
	listBoxTileset.mouseMove(x, y);
	listBoxTechTree.mouseMove(x, y);
}

void MenuStateGameSettings::render(){

	CoreData &coreData= CoreData::getInstance();
	Renderer &renderer= Renderer::getInstance();

	int i;
	Font * font= coreData.getMenuFontNormal();

	renderer.renderButton(&buttonReturn);
	renderer.renderButton(&buttonPlayNow);

	for(i=0; i<GameConstants::maxPlayers; ++i){
		renderer.renderLabel(&labelPlayers[i]);
        renderer.renderListBox(&listBoxControls[i]);
		if(listBoxControls[i].getSelectedItemIndex()!=ctClosed){
			renderer.renderListBox(&listBoxFactions[i]);
			renderer.renderListBox(&listBoxTeams[i]);
		}
    }
	renderer.renderLabel(&labelMap);
	renderer.renderLabel(&labelTileset);
	renderer.renderLabel(&labelTechTree);
	renderer.renderLabel(&labelControl);
	renderer.renderLabel(&labelFaction);
	renderer.renderLabel(&labelTeam);
	renderer.renderLabel(&labelMapInfo);

	renderer.renderListBox(&listBoxMap);
	renderer.renderListBox(&listBoxTileset);
	renderer.renderListBox(&listBoxTechTree);
}

void MenuStateGameSettings::loadGameSettings(GameSettings *gameSettings){
	gameSettings->mapPath= "maps/";
    gameSettings->mapPath+= mapFiles[listBoxMap.getSelectedItemIndex()]+".gbm";

    gameSettings->tilesetPath= "tilesets/";
    gameSettings->tilesetPath+= tilesetFiles[listBoxTileset.getSelectedItemIndex()];

    gameSettings->techPath= "techs/";
    gameSettings->techPath+= techTreeFiles[listBoxTechTree.getSelectedItemIndex()];
   
    gameSettings->factionCount=0;
    for(int i=0; i<mapInfo.players; ++i){
		ControlerType ct= static_cast<ControlerType>(listBoxControls[i].getSelectedItemIndex());
		if(ct!=ctClosed){
			if(ct==ctHuman){
				gameSettings->thisFactionIndex= gameSettings->factionCount;
			}
			gameSettings->factionControls[gameSettings->factionCount]= ct;
			gameSettings->teams[gameSettings->factionCount]= listBoxTeams[i].getSelectedItemIndex();
			gameSettings->startLocationIndex[gameSettings->factionCount]= i;
			gameSettings->factionTypesNames[gameSettings->factionCount]= factionFiles[listBoxFactions[i].getSelectedItemIndex()];
			gameSettings->factionCount++;
		}
    }
}

// ============ PRIVATE ===========================

void MenuStateGameSettings::loadMapInfo(string file, MapInfo *mapInfo){

	struct MapFileHeader{
		int32 version;
		int32 maxPlayers;
		int32 width;
		int32 height;
		int32 altFactor;
		int32 waterLevel;
		int8 title[128];
	};

	Lang &lang= Lang::getInstance();

	try{
		FILE *f= fopen(file.c_str(), "rb");
		if(f==NULL)
			throw runtime_error("Can't open file");

		MapFileHeader header;
		fread(&header, sizeof(MapFileHeader), 1, f);

		mapInfo->size.x= header.width;
		mapInfo->size.y= header.height;
		mapInfo->players= header.maxPlayers;

		mapInfo->desc= lang.get("MaxPlayers")+": "+intToStr(mapInfo->players)+"\n";
		mapInfo->desc+=lang.get("Size")+": "+intToStr(mapInfo->size.x) + " x " + intToStr(mapInfo->size.y);

		fclose(f);
	}
	catch(exception e){
		throw runtime_error("Error loading map file: "+file+'\n'+e.what());
	}

}

void MenuStateGameSettings::reloadFactions(){
	
	vector<string> results;

	findAll("techs/"+techTreeFiles[listBoxTechTree.getSelectedItemIndex()]+"/factions/*.", results);
	if(results.size()==0){
        throw runtime_error("There is no factions for this tech tree");
	}
	factionFiles= results;
   	for(int i= 0; i<results.size(); ++i){
		results[i]= formatString(results[i]);
	}
	for(int i=0; i<GameConstants::maxPlayers; ++i){
		listBoxFactions[i].setItems(results);
		listBoxFactions[i].setSelectedItemIndex(i % results.size());
    }
}

void MenuStateGameSettings::updateControlers(){
	bool humanPlayer= false;
	bool cpuPlayer= false;
	
	for(int i= 0; i<mapInfo.players; ++i){
		if(listBoxControls[i].getSelectedItemIndex() == ctHuman){
			humanPlayer= true;
		}
		if(listBoxControls[i].getSelectedItemIndex() == ctCpu || listBoxControls[i].getSelectedItemIndex() == ctCpuUltra){
			cpuPlayer= true;
		}
	}

	if( !humanPlayer || !cpuPlayer){
		listBoxControls[0].setSelectedItemIndex(ctHuman);
		listBoxControls[1].setSelectedItemIndex(ctCpu);
	}

	for(int i= mapInfo.players; i<GameConstants::maxPlayers; ++i){
		listBoxControls[i].setSelectedItemIndex(ctClosed);
	}
}

// =====================================================
// 	class MenuStateConfig
// =====================================================

MenuStateConfig::MenuStateConfig(Program *program, MainMenu *mainMenu): 
	MenuState(program, mainMenu, "config")
{
	Lang &lang= Lang::getInstance();
	Config &config= Config::getInstance();

	//create
	buttonReturn.init(200, 150, 100);	
	buttonAutoConfig.init(320, 150, 100);
	buttonOpenglInfo.init(440, 150, 100);

	//labels
	labelVolumeFx.init(200, 530);
	labelVolumeAmbient.init(200, 500);
	labelVolumeMusic.init(200, 470);

	labelLang.init(200, 400);
	
	labelFilter.init(200, 340);
	labelShadows.init(200, 310);
	labelTextures3D.init(200, 280);
	labelLights.init(200, 250);

	//list boxes
	listBoxVolumeFx.init(350, 530, 80);
	listBoxVolumeAmbient.init(350, 500, 80);
	listBoxVolumeMusic.init(350, 470, 80);
	listBoxMusicSelect.init(350, 440, 150);

	listBoxLang.init(350, 400, 170);
	
	listBoxFilter.init(350, 340, 170);
	listBoxShadows.init(350, 310, 170);
	listBoxTextures3D.init(350, 280, 80);
	listBoxLights.init(350, 250, 80);

	//set text
	buttonReturn.setText(lang.get("Return"));
	buttonAutoConfig.setText(lang.get("AutoConfig"));
	buttonOpenglInfo.setText(lang.get("GraphicInfo"));
	labelLang.setText(lang.get("Languaje"));
	labelShadows.setText(lang.get("Shadows"));
	labelFilter.setText(lang.get("Filter"));
	labelTextures3D.setText(lang.get("Textures3D"));
	labelLights.setText(lang.get("MaxLights"));
	labelVolumeFx.setText(lang.get("FxVolume"));
	labelVolumeAmbient.setText(lang.get("AmbientVolume"));
	labelVolumeMusic.setText(lang.get("MusicVolume"));

	//sound

	//lang
	vector<string> langResults;
	findAll("data/lang/*.lng", langResults);
	if(langResults.empty()){
        throw runtime_error("There is no lang file");
	}
    listBoxLang.setItems(langResults);
	listBoxLang.setSelectedItem(config.getString("Lang"));
	
	//shadows
	for(int i= 0; i<Renderer::sCount; ++i){
		listBoxShadows.pushBackItem(lang.get(Renderer::shadowsToStr(static_cast<Renderer::Shadows>(i))));
	}

	string str= config.getString("Shadows");
	Renderer::Shadows shadows = Renderer::strToShadows(str);
	listBoxShadows.setSelectedItemIndex(clamp(Renderer::strToShadows(str), 0, Renderer::sCount-1));

	//filter
	listBoxFilter.pushBackItem("Bilinear");
	listBoxFilter.pushBackItem("Trilinear");
	listBoxFilter.setSelectedItem(config.getString("Filter"));

	//textures 3d
	listBoxTextures3D.pushBackItem(lang.get("No"));
	listBoxTextures3D.pushBackItem(lang.get("Yes"));
	listBoxTextures3D.setSelectedItemIndex(clamp(config.getInt("Textures3D"), 0, 1));

	//lights
	for(int i= 1; i<=8; ++i){
		listBoxLights.pushBackItem(intToStr(i));
	}
	listBoxLights.setSelectedItemIndex(clamp(config.getInt("MaxLights")-1, 0, 7));

	//sound
	for(int i=0; i<=100; i+=5){
		listBoxVolumeFx.pushBackItem(intToStr(i));
		listBoxVolumeAmbient.pushBackItem(intToStr(i));
		listBoxVolumeMusic.pushBackItem(intToStr(i));
	}
	listBoxVolumeFx.setSelectedItem(intToStr(config.getInt("SoundVolumeFx")/5*5));
	listBoxVolumeAmbient.setSelectedItem(intToStr(config.getInt("SoundVolumeAmbient")/5*5));
	listBoxVolumeMusic.setSelectedItem(intToStr(config.getInt("SoundVolumeMusic")/5*5));
}

void MenuStateConfig::mouseClick(int x, int y, MouseButton mouseButton){

	Config &config= Config::getInstance();
	Lang &lang= Lang::getInstance();
	CoreData &coreData= CoreData::getInstance();
	SoundRenderer &soundRenderer= SoundRenderer::getInstance();

	if(buttonReturn.mouseClick(x, y)){
		soundRenderer.playFx(coreData.getClickSoundA());
		mainMenu->setState(new MenuStateRoot(program, mainMenu));
    }  
	else if(buttonAutoConfig.mouseClick(x, y)){
		soundRenderer.playFx(coreData.getClickSoundA());
		Renderer::getInstance().autoConfig();
		saveConfig();
		mainMenu->setState(new MenuStateConfig(program, mainMenu));
	}
	else if(buttonOpenglInfo.mouseClick(x, y)){
		soundRenderer.playFx(coreData.getClickSoundB());
		mainMenu->setState(new MenuStateGraphicInfo(program, mainMenu));
	}
	else if(listBoxLang.mouseClick(x, y)){
		config.setString("Lang", listBoxLang.getSelectedItem());
		lang.load("data/lang/"+config.getString("Lang"));
		saveConfig();
		mainMenu->setState(new MenuStateConfig(program, mainMenu));
		
	}
	else if(listBoxShadows.mouseClick(x, y)){
		int index= listBoxShadows.getSelectedItemIndex();
		config.setString("Shadows", Renderer::shadowsToStr(static_cast<Renderer::Shadows>(index)));
		saveConfig();
	}
	else if(listBoxFilter.mouseClick(x, y)){
		config.setString("Filter", listBoxFilter.getSelectedItem());
		saveConfig();
	}
	else if(listBoxTextures3D.mouseClick(x, y)){
		config.setInt("Textures3D", listBoxTextures3D.getSelectedItemIndex());
		saveConfig();
	}
	else if(listBoxLights.mouseClick(x, y)){
		config.setInt("MaxLights", listBoxLights.getSelectedItemIndex()+1);
		saveConfig();
	}
	else if(listBoxVolumeFx.mouseClick(x, y)){
		config.setString("SoundVolumeFx", listBoxVolumeFx.getSelectedItem());
		saveConfig();
	}
	else if(listBoxVolumeAmbient.mouseClick(x, y)){
		config.setString("SoundVolumeAmbient", listBoxVolumeAmbient.getSelectedItem());
		saveConfig();
	}
	else if(listBoxVolumeMusic.mouseClick(x, y)){
		CoreData::getInstance().getMenuMusic()->setVolume(strToInt(listBoxVolumeMusic.getSelectedItem())/100.f);
		config.setString("SoundVolumeMusic", listBoxVolumeMusic.getSelectedItem());
		saveConfig();
	}


}

void MenuStateConfig::mouseMove(int x, int y, const MouseState *ms){
	buttonReturn.mouseMove(x, y);
	buttonAutoConfig.mouseMove(x, y);
	buttonOpenglInfo.mouseMove(x, y);
	listBoxLang.mouseMove(x, y);
	listBoxVolumeFx.mouseMove(x, y);
	listBoxVolumeAmbient.mouseMove(x, y);
	listBoxVolumeMusic.mouseMove(x, y);
	listBoxLang.mouseMove(x, y);
	listBoxFilter.mouseMove(x, y);
	listBoxShadows.mouseMove(x, y);
	listBoxTextures3D.mouseMove(x, y);
	listBoxLights.mouseMove(x, y);
}

void MenuStateConfig::render(){
	Renderer &renderer= Renderer::getInstance();
	CoreData &coreData= CoreData::getInstance();

	renderer.renderButton(&buttonReturn);
	renderer.renderButton(&buttonAutoConfig);
	renderer.renderButton(&buttonOpenglInfo);
	renderer.renderListBox(&listBoxLang);
	renderer.renderListBox(&listBoxShadows);
	renderer.renderListBox(&listBoxTextures3D);
	renderer.renderListBox(&listBoxLights);
	renderer.renderListBox(&listBoxFilter);
	renderer.renderListBox(&listBoxVolumeFx);
	renderer.renderListBox(&listBoxVolumeAmbient);
	renderer.renderListBox(&listBoxVolumeMusic);
	renderer.renderLabel(&labelLang);
	renderer.renderLabel(&labelShadows);
	renderer.renderLabel(&labelTextures3D);
	renderer.renderLabel(&labelLights);
	renderer.renderLabel(&labelFilter);
	renderer.renderLabel(&labelVolumeFx);
	renderer.renderLabel(&labelVolumeAmbient);
	renderer.renderLabel(&labelVolumeMusic);
}

void MenuStateConfig::saveConfig(){
	Config &config= Config::getInstance();
	Lang &lang= Lang::getInstance();

	config.save("glest.ini");
	Renderer::getInstance().loadConfig();
	SoundRenderer::getInstance().loadConfig();
}


// =====================================================
// 	class MenuStateGraphicInfo
// =====================================================

MenuStateGraphicInfo::MenuStateGraphicInfo(Program *program, MainMenu *mainMenu): 
	MenuState(program, mainMenu, "info")
{
	buttonReturn.init(387, 100, 125);
	labelInfo.init(100, 700);
	labelMoreInfo.init(100, 500);
	labelMoreInfo.setFont(CoreData::getInstance().getMenuFontSmall());

	Renderer &renderer= Renderer::getInstance();
	glInfo= renderer.getGlInfo();
	glMoreInfo= renderer.getGlMoreInfo();
}

void MenuStateGraphicInfo::mouseClick(int x, int y, MouseButton mouseButton){
	CoreData &coreData= CoreData::getInstance();
	SoundRenderer &soundRenderer= SoundRenderer::getInstance();

	if(buttonReturn.mouseClick(x,y)){
		soundRenderer.playFx(coreData.getClickSoundA());
		mainMenu->setState(new MenuStateConfig(program, mainMenu));
    }     
}

void MenuStateGraphicInfo::mouseMove(int x, int y, const MouseState *ms){
	buttonReturn.mouseMove(x, y);
}

void MenuStateGraphicInfo::render(){
	
	CoreData &coreData= CoreData::getInstance();
	Renderer &renderer= Renderer::getInstance();
	Lang &lang= Lang::getInstance();
	Font * font= coreData.getMenuFontNormal();
	Font * fontSmall= coreData.getMenuFontSmall();
	
	buttonReturn.setText(lang.get("Return"));
	labelInfo.setText(glInfo);
	labelMoreInfo.setText(glMoreInfo);

	renderer.renderButton(&buttonReturn);
	renderer.renderLabel(&labelInfo);
	renderer.renderLabel(&labelMoreInfo);
}


// =====================================================
// 	class MenuStateAbout
// =====================================================

MenuStateAbout::MenuStateAbout(Program *program, MainMenu *mainMenu):
	MenuState(program, mainMenu, "about")
{
	Lang &lang= Lang::getInstance();
	
	//init
	buttonReturn.init(460, 100, 125);
	buttonReturn.setText(lang.get("Return"));
	
	for(int i= 0; i<aboutStringCount; ++i){
		labelAbout[i].init(280, 700-i*20);
		labelAbout[i].setText(getAboutString(i));
	}
	
	for(int i= 0; i<teammateCount; ++i){
		labelTeammateName[i].init(100+i*180, 500);
		labelTeammateRole[i].init(100+i*180, 520);
		labelTeammateName[i].setText(getTeammateName(i));
		labelTeammateRole[i].setText(getTeammateRole(i));
	}

	labelTeammateName[5].init(360, 160);
	labelTeammateRole[5].init(360, 180);
	labelTeammateName[6].init(540, 160);
	labelTeammateRole[6].init(540, 180);
}

void MenuStateAbout::mouseClick(int x, int y, MouseButton mouseButton){

	CoreData &coreData=  CoreData::getInstance();
	SoundRenderer &soundRenderer= SoundRenderer::getInstance();

	if(buttonReturn.mouseClick(x, y)){
		soundRenderer.playFx(coreData.getClickSoundA());
		mainMenu->setState(new MenuStateRoot(program, mainMenu));
    }     

}

void MenuStateAbout::mouseMove(int x, int y, const MouseState *ms){
	buttonReturn.mouseMove(x, y);
}

void MenuStateAbout::render(){
	CoreData &coreData= CoreData::getInstance();
	Renderer &renderer= Renderer::getInstance();

	renderer.renderButton(&buttonReturn);
	for(int i= 0; i<aboutStringCount; ++i){
		renderer.renderLabel(&labelAbout[i]);
	}
	for(int i= 0; i<teammateCount; ++i){
		renderer.renderLabel(&labelTeammateName[i]);
		renderer.renderLabel(&labelTeammateRole[i]);
	}
}

}}//end namespace
